module type S = sig
  type 'a m
  val run : 'a m -> 'a
  val return : 'a -> 'a m
  val bind : 'a m -> ('a -> 'b m) -> 'b m
  val callcc : (('a -> 'b m) -> 'a m) -> 'a m
end

module M : S = struct
  type r
  type 'a m = { run : ('a -> r) -> r }
  let return x = { run = fun k -> k x }
  let bind m n = { run = fun k -> m.run (fun x -> (n x).run k) }
  let callcc m =
    { run = fun k -> (m (fun a -> { run = fun _ -> k a })).run k }
  let run : type a. a m -> a = fun m ->
    let exception E of a in
      try
        ignore (m.run (fun x -> raise (E x))) ;
        assert false
      with E x -> x
end

open M
let ( let* ) = bind
let ( >>> ) m n = bind m (fun _ -> n)

let rec list_iter : type a r. (a -> unit m) -> a list -> unit m =
  fun f -> function
  | [] -> return ()
  | hd::tl -> let* _ = f hd in list_iter f tl

let find pred lst =
  callcc (fun k ->
    list_iter
      (fun x -> if pred x then k (Some x) else return ())
      lst
    >>>
    return None)

let () =
  Printf.printf "A> %d\n"
    (match M.run (find ((=) 3) [1;2;3;4])
     with Some x -> x | None -> assert false)

let find' pred lst =
  callcc (fun k ->
    list_iter
      (fun x -> if pred x then
         callcc (fun k' -> k (Some (x,k')))
       else return ())
      lst
    >>>
    return None)

let printall pred lst =
  bind (find' pred lst) (function
    | Some (x,k) -> Printf.printf "B> %d\n" x ; k ()
    | None -> return ())

let () =
  ignore (M.run (printall (fun x -> x mod 2 = 1) [1;2;3;4]))
