module type MONAD = sig
  type 'a m
  val return : 'a -> 'a m
  val bind : 'a m -> ('a -> 'b m) -> 'b m
end

module type COUNTERMONAD = sig
  include MONAD
  val get : int m
  val set : int -> unit m
  val incr : unit m
  val run : 'a m -> 'a
end

module Counter : COUNTERMONAD = struct

  type 'a m = int -> 'a*int

  let return x = fun i -> x,i
  let bind f g = fun i -> let x,j = f i in g x j

  let run f = fst (f 0)

  let get = fun i -> i,i
  let incr = fun i -> (),i+1
  let set x = fun i -> (),x

end

(** Using our monad *)

let i =
  let open Counter in
  run
    (bind get (fun x ->
    (bind incr (fun _ ->
    (bind (set x) (fun _ ->
    (bind incr (fun _ ->
    get))))))))

let () = Printf.printf "> %d\n" i

(** Syntactic sugar *)

module NiceCounter = struct
  include Counter
  let (>>=) = bind
  let (let*) = bind
  let (>>>) f g = let* _ = f in g
end

let i =
  let open NiceCounter in
  run begin
    let* x = get in
    incr >>>
    set x >>>
    incr >>>
    get
  end

let () = Printf.printf "> %d\n" i
